"""
The following function was copied and modified from the porespy project (distributed under the MIT License).
See https://github.com/PMEAL/porespy/blob/dev/porespy/generators/_imgen.py for more information.
See https://github.com/PMEAL/porespy/blob/dev/LICENSE for the license.
"""

from scipy.ndimage import distance_transform_edt as edt
from skimage.segmentation import watershed
import numpy as np
from pumapy.utilities.workspace import Workspace
from pumapy.utilities.logger import print_warning
from pumapy.filters.filters import filter_gaussian
import sys


def generate_random_fibers(shape, radius, nfibers=None, porosity=None, phi=90., theta=90., length=None, max_iter=3,
                           allow_intersect=True, segmented=True):
    """ Generates random fibers from number of fibers or porosity

        :param shape: the shape of the workspace to generate in (Nx, Ny, Nz) where N is the number of voxels.
        :type shape: (int, int, int)
        :param radius: the radius of the fibers in voxels
        :type radius: int
        :param nfibers: the number of fibers to add to the domain. Adjust this value to control the final porosity,
            which is not easily specified since cylinders overlap and intersect different fractions of the domain
        :type nfibers: int
        :param porosity: the target value for the porosity of the generated mat. The function uses an
            algorithm for predicting the number of required number of cylinder, and refines this over a certain number of
            fractional insertions (according to the 'iterations' input)
        :type porosity: float
        :param phi: a value between 0 and 90 that controls the amount that the fibers lie *out of* the XY plane,
            with 0 meaning all fibers lie in the XY plane, and 90 meaning that cylinders are randomly oriented out of the
            plane by as much as +/- 90 degrees
        :type phi: float
        :param theta: a value between 0 and 90 that controls the amount of rotation *in the* XY plane,
            with 0 meaning all fibers point in the X-direction, and 90 meaning they are randomly rotated about the
            Z axis by as much as +/- 90 degrees
        :type theta: float
        :param length: the length of the cylinders to add.  If ``None`` (default) then the cylinders will extend beyond the
            domain in both directions so no ends will exist. If a scalar value is given it will be interpreted as the
            Euclidean distance between the two ends of the cylinder.  Note that one or both of the ends *may* still lie
            outside the domain, depending on the randomly chosen center point of the cylinder
        :type length: float
        :param max_iter: the number of fractional fiber insertions used to target the requested porosity. By default a
            value of 3 is used (and this is typically effective in getting very close to the targeted porosity),
            but a greater number can be input to improve the achieved porosity
        :type max_iter: int
        :param allow_intersect: allow intersection betweem the fibers
        :type allow_intersect: bool
        :param segmented: return a domain that is already segmented (i.e. each fiber with unique ID) or
            with grayscales 0-255 with threshold=128 for the input diameter
        :type segmented: bool
        :return: random fibers domain
        :rtype: Workspace

        :Example:
        >>> import pumapy as puma
        >>> # specify porosity
        >>> ws_fibers = puma.generate_random_fibers(shape=(100, 100, 100), radius=4, porosity=0.8, phi=90, theta=90, length=200, allow_intersect=True, segmented=True)
        >>> puma.render_volume(ws_fibers, cutoff=(1, ws_fibers.max()), cmap='jet')
        >>> puma.render_orientation(ws_fibers)
        >>> # specify number of fibers
        >>> ws_fibers = puma.generate_random_fibers(shape=(100, 100, 100), radius=4, nfibers=100, phi=90, theta=90, length=200, allow_intersect=True, segmented=False)
        >>> puma.render_volume(ws_fibers, cutoff=(1, ws_fibers.max()), cmap='jet')
        >>> puma.render_orientation(ws_fibers)
        >>> # don't allow intersection between fibers
        >>> ws_fibers = puma.generate_random_fibers(shape=(100, 100, 100), radius=4, porosity=0.8, phi=90, theta=90, length=200, allow_intersect=False, segmented=True)
        >>> puma.render_volume(ws_fibers, cutoff=(1, ws_fibers.max()), cmap='jet')
        >>> puma.render_orientation(ws_fibers)
    """

    # error checks and warnings
    if max_iter < 3:
        raise Exception("Iterations must be >= 3")
    if nfibers is None and porosity is None:
        raise Exception("'nfibers' and 'porosity' can't be both None")
    if nfibers is not None and porosity is not None:
        print_warning("'nfibers' and 'porosity' have both been provided: only nfibers used.")

    # logging
    ws = Workspace()
    ws.log.log_section("Generating Random Fiber Domain")
    ws.log.log_value("Domain Shape: ", shape)
    ws.log.log_value("Fiber radius: ", radius)
    ws.log.log_line(f"Porosity: {porosity}" if nfibers is None else f"Number of fibers: {nfibers}")
    ws.log.log_line(f"Phi and theta angle variations: ({phi}, {theta})")
    ws.log.log_line(f"Fiber length: {length}")
    ws.log.log_line(f"Allow intersections: {allow_intersect}")
    ws.log.write_log()

    ids_orient_dict = {}
    fiber_count = 0

    img = np.zeros(shape, dtype=int)

    # run solver for provided number of fibers
    if nfibers is not None:
        if allow_intersect:
            img, _ = _generate_fibers(shape=shape, radius=radius, nfibers=nfibers, n_fibers_total=nfibers,
                                      fiber_count=fiber_count, ids_orient_dict=ids_orient_dict,
                                      allow_intersect=allow_intersect, phi=phi, theta=theta, length=length)
        else:
            _generate_fibers(shape=shape, radius=radius, nfibers=nfibers, n_fibers_total=nfibers,
                             fiber_count=fiber_count, ids_orient_dict=ids_orient_dict,
                             allow_intersect=allow_intersect, phi=phi, theta=theta, length=length, existing_img=img)

    else:  # porosity provided
        vol_total = float(np.prod(shape))

        length_estimate = vol_total ** (1 / 3) if length is None else length

        vol_fiber = length_estimate * np.pi * radius * radius
        n_pixels_to_add = -np.log(porosity) * vol_total

        subdif = 0.8 / np.sum(np.arange(1, max_iter) ** 2)
        fractions = [0.2]
        for i in range(1, max_iter):
            fractions.append(fractions[i - 1] + (max_iter - i) ** 2 * subdif)

        img = np.zeros(shape, dtype=int)
        current_porosity = 1.
        for i, frac in enumerate(fractions):
            n_fibers_total = n_pixels_to_add / vol_fiber
            n_fibers = int(np.ceil(frac * n_fibers_total) - fiber_count)
            if current_porosity < porosity:
                break

            if allow_intersect:
                new_fibers, mask = _generate_fibers(shape, radius, nfibers=n_fibers, n_fibers_total=int(round(n_fibers_total)),
                                                    fiber_count=fiber_count, ids_orient_dict=ids_orient_dict, allow_intersect=allow_intersect,
                                                    phi=phi, theta=theta, length=length)
                new_fibers = new_fibers[mask]
                img[mask] = new_fibers
            else:
                _generate_fibers(shape, radius, nfibers=n_fibers,
                                 n_fibers_total=int(round(n_fibers_total)),
                                 fiber_count=fiber_count, ids_orient_dict=ids_orient_dict,
                                 allow_intersect=allow_intersect,
                                 phi=phi, theta=theta, length=length, existing_img=img)

            fiber_count = np.unique(img).shape[0]

            if i != max_iter - 1:  # update target for next iteration
                porosity = np.sum(img == 0) / img.size
                vol_added = -np.log(porosity) * vol_total
                vol_fiber = vol_added / fiber_count

    ws.matrix = img.astype(np.uint16)

    print("\nAssigning orientations ... ", end='')
    k = np.array(list(ids_orient_dict.keys()))
    v = np.array(list(ids_orient_dict.values()))
    mapping_ar = np.zeros((k.max() + 1, 3), dtype=v.dtype)
    mapping_ar[k] = v
    ws.orientation = mapping_ar[ws.matrix]
    print("Done")

    if not segmented:  # add grayscale gradients with Gaussian
        ws.binarize(1)
        ws.matrix *= 255
        filter_gaussian(ws, sigma=int(round((radius//2-1))))
        porosity = ws.porosity(cutoff=(0, 127))
    else:
        porosity = ws.porosity()
    print("Generated random fibers domain with porosity: {}".format(porosity))
    ws.log.log_section("Finished Random Fiber Generation")
    ws.log.write_log()
    return ws


def _generate_fibers(shape, radius, nfibers, n_fibers_total, fiber_count, ids_orient_dict, allow_intersect,
                     phi, theta, length=None, existing_img=None):
    """ Generates random fibers given nfibers"""

    # error checks
    shape = np.array(shape)
    if np.size(shape) == 1:
        shape = np.full((3, ), int(shape))
    elif np.size(shape) == 2:
        raise Exception("Shape can only be 3D")

    H = np.sqrt(np.sum(np.square(shape))).astype(int)
    if length is None:
        length = 2 * H
    R = min(int(length / 2), 2 * H)

    if (phi > 90) or (phi < 0):
        raise Exception('phi_max must be between 0 and 90')
    if (theta > 90) or (theta < 0):
        raise Exception('theta_max must be between 0 and 90')

    n = 0  # current iteration number of fibers
    L = min(H, R)

    if not allow_intersect:
        mask = np.ones(shape, dtype=bool)
    else:
        im = np.zeros(shape, dtype=int)

    # maximum unsuccessful attempts of inserting a fiber
    max_attempts = 100000
    attempt_num = 0

    while n < nfibers:
        if attempt_num > max_attempts:
            raise Exception(f"While loop did not converge after {max_attempts} iterations.")
        else:
            attempt_num += 1

        x = np.random.rand(3) * (shape + 2 * L)

        phi_rand = (np.pi / 2 - np.pi * np.random.rand()) * phi / 90
        theta_rand = (np.pi / 2 - np.pi * np.random.rand()) * theta / 90
        unit_vector = np.array([np.cos(phi_rand) * np.cos(theta_rand),
                                np.cos(phi_rand) * np.sin(theta_rand),
                                np.sin(phi_rand)])
        x0 = R * unit_vector
        [x0, x1] = [x + x0, x - x0]

        x0 = np.around(x0).astype(int)
        x1 = np.around(x1).astype(int)
        L = np.amax(np.absolute([[x1[0] - x0[0]], [x1[1] - x0[1]], [x1[2] - x0[2]]])) + 1
        crds = [np.rint(np.linspace(x0[0], x1[0], L)).astype(int),
                np.rint(np.linspace(x0[1], x1[1], L)).astype(int),
                np.rint(np.linspace(x0[2], x1[2], L)).astype(int)]

        lower = ~np.any(np.vstack(crds).T < [L, L, L], axis=1)
        upper = ~np.any(np.vstack(crds).T >= shape + L, axis=1)
        valid = upper * lower
        if np.any(valid):
            if not allow_intersect:
                mask[crds[0][valid] - L, crds[1][valid] - L, crds[2][valid] - L] = 0
                mask = edt(mask) < radius
                if existing_img[mask].max() == 0:
                    fiber_count += 1
                    existing_img[mask] = fiber_count
                    ids_orient_dict[fiber_count] = unit_vector
                    n += 1
                    attempt_num = 0
                mask.fill(1)
            else:
                fiber_count += 1
                ids_orient_dict[fiber_count] = unit_vector
                im[crds[0][valid] - L, crds[1][valid] - L, crds[2][valid] - L] = fiber_count
                n += 1
                attempt_num = 0  # restart counter if change happens
            sys.stdout.write("\rFibers created: {} --> approximate target: {} ".format(fiber_count, n_fibers_total))

    if allow_intersect:
        distance_map = edt(im == 0)
        mask = distance_map < radius
        return watershed(distance_map, im, mask=mask), mask



from pumapy.utilities.workspace import Workspace
from pumapy.utilities.isosurface import generate_isosurface
from pumapy.utilities.logger import print_warning
import pyvista as pv
import numpy as np
import string
import random


def render_volume(workspace, cutoff=None, solid_color=None, style='surface', origin=(0., 0., 0.),
                  window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True, plot_directly=True,
                  show_axes=True, show_outline=True, cmap='gray', add_to_plot=None, notebook=False):
    """ Volume render using Pyvista Threshold filter

        :param workspace: domain
        :type workspace: Workspace or np.ndarray
        :param cutoff: specifying the values to render
        :type cutoff: (int, int) or None
        :param solid_color: if set to None (default), the material is colored by the matrix's values.
                            Otherwise, a solid color can be specified (e.g. for white (1., 1., 1.))
        :type solid_color: (float, float, float) or None
        :param style: specifying the representation style ('surface', 'edges', 'wireframe', 'points')
        :type style: string
        :param origin: origin of the data as
        :type origin: (float, float, float)
        :param window_size: with the popup window size
        :type window_size: (int, int)
        :param opacity: opacity of volume
        :type opacity: float
        :param background: color of the background from (0., 0., 0.) (black) to (1., 1., 1.) (white)
        :type: background: (float, float, float)
        :param show_grid: show the grid with the size of the sides
        :type show_grid: bool
        :param plot_directly: whether to return a Plotter object (to make further changes to it) or show the plot directly
        :type plot_directly: bool
        :param show_axes: show orientation axis in the bottom left corner
        :type show_axes: bool
        :param show_outline: show the bounding box outline of the domain
        :type show_outline: bool
        :param cmap: matplotlib colormap to use (overwritten by solid_color if specified)
        :type cmap: str
        :param add_to_plot: pass an already existing plotter object to add on top of this plot
        :type add_to_plot: pyvista.Plotter or None
        :param notebook: plotting interactively in a jupyter notebook (overwrites show_grid to False)
        :type notebook: bool
        :return: None is plot_directly is True, otherwise a plotter object
        :rtype: pyvista.Plotter or None

        :Example:
        >>> import pumapy as puma
        >>> ws_volume = puma.import_3Dtiff(puma.path_to_example_file("200_fiberform.tif"), 1.3e-6)
        >>> puma.render_volume(ws_volume)
    """
    r = Renderer(existing_plot=add_to_plot, filter_type="threshold", workspace=workspace, cutoff=cutoff,
                 solid_color=solid_color, style=style, origin=origin, window_size=window_size, opacity=opacity,
                 background=background, show_grid=show_grid, plot_directly=plot_directly, show_axes=show_axes,
                 show_outline=show_outline, cmap=cmap, notebook=notebook)

    return r.render()


def render_contour(workspace, cutoff, solid_color=(1., 1., 1.), style='surface', origin=(0., 0., 0.),
                   window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True, plot_directly=True,
                   show_axes=True, show_outline=True, add_to_plot=None, notebook=False):
    """ Contour render using Pyvista Contour filter

        :param workspace: domain
        :type workspace: Workspace or np.ndarray
        :param cutoff: specifying the values at which the isosurface is created
        :type cutoff: (int, int)
        :param solid_color: a solid color to color the surface (e.g. for white (1., 1., 1.))
        :type solid_color: (float, float, float)
        :param style: specifying the representation style ('surface', 'edges', 'wireframe', 'points')
        :type style: string
        :param origin: origin of the data as
        :type origin: (float, float, float)
        :param window_size: with the popup window size
        :type window_size: (int, int)
        :param opacity: opacity of contour
        :type opacity: float
        :param background: color of the background from (0., 0., 0.) (black) to (1., 1., 1.) (white)
        :type: background: (float, float, float)
        :param show_grid: show the grid with the size of the sides
        :type show_grid: bool
        :param plot_directly: whether to return a Plotter object (to make further changes to it) or show the plot directly
        :type plot_directly: bool
        :param show_axes: show orientation axis in the bottom left corner
        :type show_axes: bool
        :param show_outline: show the bounding box outline of the domain
        :type show_outline: bool
        :param add_to_plot: pass an already existing plotter object to add on top of this plot
        :type add_to_plot: pyvista.Plotter or None
        :param notebook: plotting interactively in a jupyter notebook (overwrites show_grid to False)
        :type notebook: bool
        :return: None is plot_directly is True, otherwise a plotter object
        :rtype: pyvista.Plotter or None

        :Example:
        >>> import pumapy as puma
        >>> ws_contour = puma.import_3Dtiff(puma.path_to_example_file("50_artfibers.tif"))
        >>> puma.render_contour(ws_contour, (128,255))
    """
    r = Renderer(existing_plot=add_to_plot, filter_type="contour", workspace=workspace, cutoff=cutoff,
                 solid_color=solid_color, style=style, origin=origin, window_size=window_size, opacity=opacity,
                 background=background, show_grid=show_grid, plot_directly=plot_directly, show_axes=show_axes,
                 show_outline=show_outline, notebook=notebook)

    return r.render()


def render_orientation(workspace, scale_factor=1., solid_color=None, style='surface', origin=(0., 0., 0.),
                       window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True, cmap=None,
                       plot_directly=True, show_axes=True, show_outline=True, add_to_plot=None, notebook=False,
                       sampling=None):
    """ Orientation render using Pyvista Glyph filter

        :param workspace: domain
        :type workspace: Workspace or np.ndarray
        :param scale_factor: scale the arrows by a factor
        :type scale_factor: float
        :param solid_color: a solid color for the arrows. Deafult is None, which colors the vectors by their magnitude.
                            To color white, input set solid_color=(1., 1., 1.)
        :type solid_color: None or (float, float, float)
        :param style: specifying the representation style ('surface', 'edges', 'wireframe', 'points')
        :type style: string
        :param origin: origin of the data as
        :type origin: (float, float, float)
        :param window_size: with the popup window size
        :type window_size: (int, int)
        :param opacity: opacity of arrows
        :type opacity: float
        :param background: color of the background from (0., 0., 0.) (black) to (1., 1., 1.) (white)
        :type: background: (float, float, float)
        :param show_grid: show the grid with the size of the sides
        :type show_grid: bool
        :param cmap: matplotlib colormap to use (overwritten by solid_color if specified)
        :type cmap: str or None
        :param plot_directly: whether to return a Plotter object (to make further changes to it) or show the plot directly
        :type plot_directly: bool
        :param show_axes: show orientation axis in the bottom left corner
        :type show_axes: bool
        :param show_outline: show the bounding box outline of the domain
        :type show_outline: bool
        :param add_to_plot: pass an already existing plotter object to add on top of this plot
        :type add_to_plot: pyvista.Plotter or None
        :param notebook: plotting interactively in a jupyter notebook (overwrites show_grid to False)
        :type notebook: bool
        :param sampling: number of arrows to sample (None means plot an arrow at each voxel)
        :type sampling: int or None
        :return: None is plot_directly is True, otherwise a plotter object
        :rtype: pyvista.Plotter or None

        :Example:
        >>> import pumapy as puma
        >>> ws_orientation = puma.import_3Dtiff(puma.path_to_example_file("100_fiberform.tif"), 1.3e-6)
        >>> puma.compute_orientation_st(ws_orientation, (90, 255))
        >>> puma.render_orientation(ws_orientation)
    """
    r = Renderer(existing_plot=add_to_plot, filter_type="glyph", workspace=workspace, solid_color=solid_color, style=style,
                 origin=origin, window_size=window_size, opacity=opacity, background=background, show_grid=show_grid,
                 plot_directly=plot_directly, show_axes=show_axes, show_outline=show_outline, cmap=cmap,
                 scale_factor=scale_factor, notebook=notebook, sampling=sampling)

    return r.render()


def render_warp(workspace, scale_factor=1., color_by='magnitude', style='surface', origin=(0., 0., 0.),
                window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True, cmap='jet',
                plot_directly=True, show_axes=True, show_outline=True, add_to_plot=None, notebook=False):
    """ Warp a domain by an orientation field

        :param workspace: domain with both matrix and orientation defined
        :type workspace: Workspace
        :param scale_factor: scale the orientation by a factor
        :type scale_factor: float
        :param color_by: it can be 'magnitude', which colors warped domain by magnitude of vectors, or 'x', 'y' or 'z',
                         which colors it by the orientation component, or 'matrix', which colors it by the matrix phases
        :type color_by: str
        :param style: specifying the representation style ('surface', 'edges', 'wireframe', 'points')
        :type style: string
        :param origin: origin of the data as
        :type origin: (float, float, float)
        :param window_size: with the popup window size
        :type window_size: (int, int)
        :param opacity: opacity of arrows
        :type opacity: float
        :param background: color of the background from (0., 0., 0.) (black) to (1., 1., 1.) (white)
        :type: background: (float, float, float)
        :param show_grid: show the grid with the size of the sides
        :type show_grid: bool
        :param cmap: matplotlib colormap to use (overwritten by solid_color if specified)
        :type cmap: str
        :param plot_directly: whether to return a Plotter object (to make further changes to it) or show the plot directly
        :type plot_directly: bool
        :param show_axes: show orientation axis in the bottom left corner
        :type show_axes: bool
        :param show_outline: show the bounding box outline of the domain
        :type show_outline: bool
        :param add_to_plot: pass an already existing plotter object to add on top of this plot
        :type add_to_plot: pyvista.Plotter or None
        :param notebook: plotting interactively in a jupyter notebook (overwrites show_grid to False)
        :type notebook: bool
        :return: None is plot_directly is True, otherwise a plotter object
        :rtype: pyvista.Plotter or None

        :Example:
        >>> import pumapy as puma
        >>> ws = puma.Workspace.from_shape_value((20, 25, 18), 1)
        >>> ws[ws.matrix.shape[0]//2:] = 2
        >>> elast_map = puma.ElasticityMap()
        >>> elast_map.add_isotropic_material((1, 1), 200, 0.3)
        >>> elast_map.add_isotropic_material((2, 2), 400, 0.1)
        >>> bc = puma.ElasticityBC(ws)
        >>> bc.dirichlet[0] = 0  # hold x -ve face
        >>> bc.dirichlet[-1, :, :, 0] = 10   # displace x +ve face by 1 in x direction
        >>> bc.dirichlet[-1, :, :, 1:] = 0  # hold x +ve face in y and z directions
        >>> ws.orientation, _, _ = puma.compute_stress_analysis(ws, elast_map, bc, side_bc='f', solver_type="direct")
        >>> puma.render_warp(ws, color_by='y', style='edges')
    """
    if not isinstance(workspace, Workspace):
        raise Exception('Input is not a pumapy.Workspace.')
    if color_by == "matrix":
        if (workspace.matrix.shape[0] != workspace.orientation.shape[0] or
                workspace.matrix.shape[1] != workspace.orientation.shape[1] or
                workspace.matrix.shape[2] != workspace.orientation.shape[2]):
            raise Exception('Workspace has to have matrix and orientation variables of the same shape.')

    r = Renderer(existing_plot=add_to_plot, filter_type="warp", workspace=workspace, solid_color=color_by, style=style,
                 origin=origin, window_size=window_size, opacity=opacity, background=background, show_grid=show_grid,
                 plot_directly=plot_directly, show_axes=show_axes, show_outline=show_outline, cmap=cmap,
                 scale_factor=scale_factor, notebook=notebook)

    return r.render()


def render_contour_multiphase(workspace, cutoffs, solid_colors=None, style='surface', origin=(0., 0., 0.),
                              window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True,
                              plot_directly=True, show_axes=True, show_outline=True, add_to_plot=None, notebook=False):
    """ Contour render for multi-phase materials using Pyvista

        :param workspace: domain
        :type workspace: Workspace or ndarray
        :param cutoffs: n cutoffs is the number of materials. specifies the low and high cutoff ranges
        :type cutoffs: tuple
        :param solid_colors: solid colors to color the different phases' surface e.g. for white ((1., 1., 1.), (0., 0., 0.), ...)
        :type solid_colors: tuple or None
        :param style: specifying the representation style ('surface', 'edges', 'wireframe', 'points')
        :type style: string
        :param origin: origin of the data as
        :type origin: (float, float, float)
        :param window_size: with the popup window size
        :type window_size: (int, int)
        :param opacity: opacity of contour
        :type opacity: float
        :param background: color of the background from (0., 0., 0.) (black) to (1., 1., 1.) (white)
        :type: background: (float, float, float)
        :param show_grid: show the grid with the size of the sides
        :type show_grid: bool
        :param plot_directly: whether to return a Plotter object (to make further changes to it) or show the plot directly
        :type plot_directly: bool
        :param show_axes: show orientation axis in the bottom left corner
        :type show_axes: bool
        :param show_outline: show the bounding box outline of the domain
        :type show_outline: bool
        :param add_to_plot: pass an already existing plotter object to add on top of this plot
        :type add_to_plot: pyvista.Plotter or None
        :param notebook: plotting interactively in a jupyter notebook (overwrites show_grid to False)
        :type notebook: bool
        :return: None is plot_directly is True, otherwise a plotter object
        :rtype: pyvista.Plotter or None

        :Example:
        >>> import pumapy as puma
        >>> ws_multiphase = puma.import_3Dtiff(puma.path_to_example_file("100_fiberform.tif"), 1.3e-6)
        >>> puma.render_contour_multiphase(ws_multiphase, ((100, 150), (150, 255)))
    """

    if add_to_plot is None:
        p = pv.Plotter(notebook=notebook)
    else:
        p = add_to_plot
    if notebook:  # observed problems in rendering with grid in jupyter
        show_grid = False
    for i, cutoff in enumerate(cutoffs):
        if solid_colors is None or len(solid_colors) != len(cutoffs):
            solid_color = tuple(np.random.rand(3))
        else:
            solid_color = solid_colors[i]
        r = Renderer(existing_plot=p, filter_type="contour", workspace=workspace, cutoff=cutoff, solid_color=solid_color,
                     style=style, origin=origin, window_size=window_size, opacity=opacity, background=background,
                     show_grid=show_grid, plot_directly=False, show_axes=show_axes, show_outline=show_outline)

        p = r.render()
    if plot_directly:
        p.show()
    else:
        return p


class Renderer:
    def __init__(self, existing_plot=None, filter_type=None, workspace=None, cutoff=None, solid_color=None, style=None,
                 origin=None, window_size=None, opacity=None, background=None, show_grid=None, plot_directly=None,
                 show_axes=None, show_outline=None, cmap=None, scale_factor=None, notebook=None, sampling=None):
        self.filter_type = filter_type
        self.cutoff = cutoff
        self.solid_color = solid_color
        self.style = style
        self.origin = origin
        self.window_size = window_size
        self.opacity = opacity
        self.background = background
        self.show_grid = show_grid
        self.plot_directly = plot_directly
        self.show_axes = show_axes
        self.show_outline = show_outline
        self.cmap = cmap
        self.voxel_length = 1
        self.scale_factor = scale_factor

        if isinstance(workspace, Workspace):
            self.voxel_length = workspace.voxel_length

            if self.filter_type == "glyph":
                self.scale_factor *= self.voxel_length
                # reducing number of arrows by sampling
                if sampling is not None and sampling >= np.prod(workspace.orientation.shape[:3]):
                    print_warning("Sampling has to be less than the number of voxels. Setting it to None.")
                    sampling = None
                if sampling is not None:
                    mag = workspace.orientation_magnitude()
                    inds = np.where(mag > 0.5)
                    rand_choice = np.random.choice(inds[0].shape[0], sampling, replace=False)
                    inds = (inds[0][rand_choice], inds[1][rand_choice], inds[2][rand_choice])
                    self.array = np.zeros_like(workspace.orientation)
                    self.array[inds[0], inds[1], inds[2]] = workspace.orientation[inds[0], inds[1], inds[2]]
                else:
                    self.array = workspace.orientation

            elif self.filter_type == "warp":
                self.array = workspace.matrix
                self.orientation = workspace.orientation
                # self.scale_factor *= self.voxel_length
            else:
                self.array = workspace.matrix

        elif isinstance(workspace, np.ndarray):
            if self.filter_type == "glyph":
                if not (workspace.ndim == 4 and workspace.shape[3] == 3):
                    raise Exception("For rendering orientation, need a numpy.ndarray of shape (X, Y, Z, 3)")
            self.array = workspace
        else:
            raise Exception("Need to input either pumapy.Workspace or a numpy.ndarray")

        if self.filter_type != "contour" and self.filter_type != "warp":
            self.grid = pv.UniformGrid()
            self.grid.origin = self.origin
        elif self.filter_type == "warp":
            x = np.linspace(0, self.orientation.shape[0] - 1, self.orientation.shape[0])
            y = np.linspace(0, self.orientation.shape[1] - 1, self.orientation.shape[1])
            z = np.linspace(0, self.orientation.shape[2] - 1, self.orientation.shape[2])
            x, y, z = np.meshgrid(x, y, z, indexing='ij')
            self.grid = pv.StructuredGrid(x, y, z)

        self.filter = None
        if existing_plot is None:
            self.p = pv.Plotter(notebook=notebook)
        else:
            self.p = existing_plot

        if notebook:  # observed problems in rendering with grid in jupyter
            self.show_grid = False

    def render(self):
        self.build_plotter()

        if self.plot_directly:
            self.p.show()
        else:
            return self.p

    def build_plotter(self):
        if self.style == 'edges':
            show_edges = True
            self.style = 'surface'
        else:
            show_edges = False

        if self.filter_type == "threshold":
            self.grid.dimensions = np.array(self.array.shape) + 1
            self.grid.spacing = (self.voxel_length, self.voxel_length, self.voxel_length)
            self.grid.cell_data["values"] = self.array.flatten(order="F")
            self.filter = self.grid.threshold(self.cutoff)

        elif self.filter_type == "contour":
            tri_mesh = generate_isosurface(self.array, self.cutoff)
            tri_mesh.verts += self.origin
            tri_mesh.create_mesh()
            self.filter = tri_mesh.mesh

        elif self.filter_type == "glyph":
            self.grid.dimensions = np.array(self.array.shape[:3]) + 1
            self.grid.spacing = (self.voxel_length, self.voxel_length, self.voxel_length)
            tmp = np.zeros((self.array[:, :, :, 0].size, 3), dtype=float)
            for i in range(3):
                tmp[:, i] = self.array[:, :, :, i].ravel(order='F')
            self.grid.cell_data["scalars"] = np.linalg.norm(self.array, axis=3).ravel(order='F')
            self.grid.cell_data["vectors"] = tmp
            self.filter = self.grid.glyph(orient="vectors", scale="scalars", factor=self.scale_factor, geom=pv.Arrow())

        elif self.filter_type == "warp":
            tmp = np.zeros((self.orientation[:, :, :, 0].size, 3), dtype=float)
            for i in [0, 1, 2]:
                tmp[:, i] = self.orientation[:, :, :, i].ravel(order='F')
            self.grid['vectors'] = tmp
            if self.solid_color == 'magnitude':
                self.grid['scalars'] = np.linalg.norm(self.orientation, axis=3).ravel(order='F')
            elif self.solid_color == 'x':
                self.grid['scalars'] = self.orientation[:, :, :, 0].flatten(order="F")
            elif self.solid_color == 'y':
                self.grid['scalars'] = self.orientation[:, :, :, 1].flatten(order="F")
            elif self.solid_color == 'z':
                self.grid['scalars'] = self.orientation[:, :, :, 2].flatten(order="F")
            else:  # defaulting to matrix, even if not correct
                if self.solid_color != 'matrix':
                    print_warning("Options for color_by input are: 'magnitude', 'x', 'y', 'z', 'matrix'")
                self.grid['scalars'] = self.array.flatten(order="F")
            self.filter = self.grid.warp_by_vector('vectors', factor=self.scale_factor)

        if self.filter_type != "warp":
            self.p.add_mesh(self.filter, color=self.solid_color, show_edges=show_edges, style=self.style,
                            cmap=self.cmap, show_scalar_bar=False, opacity=self.opacity)
        else:
            self.p.add_mesh(self.filter, scalars='scalars', interpolate_before_map=False, show_edges=show_edges,
                            style=self.style, cmap=self.cmap, show_scalar_bar=False, opacity=self.opacity)

        # this is because warp uses solid_color to select what to color the domain by
        if self.solid_color is None or isinstance(self.solid_color, str):
            # the title has to be a unique string for each subplot
            self.p.add_scalar_bar(title="".join([random.choice(string.ascii_letters) for _ in range(10)]),
                                  title_font_size=1, height=0.5, vertical=True, position_x=0.05, position_y=0.3)

        self.p.background_color = self.background
        self.p.window_size = self.window_size
        if self.show_outline:
            self.p.add_bounding_box(color="k")
        if self.show_axes:
            self.p.show_axes()
        if self.show_grid:
            self.p.show_grid()



from pumapy.physicsmodels.isotropic_conductivity import IsotropicConductivity
from pumapy.physicsmodels.mpfa_conductivity import AnisotropicConductivity
from pumapy.physicsmodels.property_maps import IsotropicConductivityMap, AnisotropicConductivityMap


def compute_thermal_conductivity(workspace, cond_map, direction, side_bc='s', prescribed_bc=None, tolerance=1e-4,
                                 maxiter=10000, solver_type='bicgstab', display_iter=True, print_matrices=(0, 0, 0, 0, 0)):
    """ Compute the thermal conductivity

        :param workspace: domain
        :type workspace: pumapy.Workspace
        :param cond_map: local constituents themal conductivities
        :type cond_map: IsotropicConductivityMap or AnisotropicConductivityMap
        :param direction: direction for solve ('x','y', or 'z')
        :type direction: string
        :param side_bc: side boundary conditions can be symmetric ('s'), periodic ('p') or dirichlet ('d')
        :type side_bc: string
        :param prescribed_bc: 3D array holding dirichlet BC.
        :type prescribed_bc: pumapy.ConductivityBC or None
        :param tolerance: tolerance for iterative solver
        :type tolerance: float
        :param maxiter: maximum Iterations for solver
        :type maxiter: int
        :param solver_type: solver type, options: 'bicgstab' (default), 'cg', 'gmres', 'direct'
        :type solver_type: string
        :param display_iter: display iterations and residual
        :type display_iter: bool
        :param print_matrices: corresponding to b, E, A, T, q decimal places. If 0, they are not printed
        :type print_matrices: (bool, bool, bool, bool, bool)
        :return: thermal conductivity, temperature field, flux
        :rtype: ((float, float, float), numpy.ndarray, numpy.ndarray)

        :Example:
        >>> import pumapy as puma
        >>> ws_fiberform = puma.import_3Dtiff(puma.path_to_example_file("200_fiberform.tif"), 1.3e-6)
        >>> ws_fiberform.rescale(0.5, segmented=False)
        >>>
        >>> # Conductivity with Isotropic local phases
        >>> cond_map = puma.IsotropicConductivityMap()
        >>> cond_map.add_material((0, 89), 0.0257)
        >>> cond_map.add_material((90, 255), 12)
        >>> k_eff_x, T_x, q_x = puma.compute_thermal_conductivity(ws_fiberform, cond_map, 'x', 's')
        >>>
        >>> # Conductivity with Anisotropic local phases
        >>> puma.compute_orientation_st(ws_fiberform, cutoff=(90, 255))
        >>> cond_map = puma.AnisotropicConductivityMap()
        >>> # conductivity of the void phase to be 0.0257 (air at STP)
        >>> cond_map.add_isotropic_material((0, 89), 0.0257)
        >>> # axial fiber conductivity of 12, radial fiber conductivity of 0.7
        >>> cond_map.add_material_to_orient((90, 255), 12., 0.7)
        >>> k_eff_z, T_z, q_z = puma.compute_thermal_conductivity(ws_fiberform, cond_map, 'z', 's')
    """
    if isinstance(cond_map, IsotropicConductivityMap):
        solver = IsotropicConductivity(workspace, cond_map, direction, side_bc, prescribed_bc, tolerance, maxiter,
                                       solver_type, display_iter)
    elif isinstance(cond_map, AnisotropicConductivityMap):
        solver = AnisotropicConductivity(workspace, cond_map, direction, side_bc, prescribed_bc, tolerance, maxiter,
                                         solver_type, display_iter, print_matrices)
    else:
        raise Exception("cond_map has to be an IsotropicConductivityMap or AnisotropicConductivityMap")

    solver.error_check()

    solver.log_input()
    solver.compute()
    solver.log_output()
    return solver.keff, solver.T, solver.q


def compute_electrical_conductivity(workspace, cond_map, direction, side_bc='p', prescribed_bc=None, tolerance=1e-4,
                                    maxiter=10000, solver_type='bicgstab', display_iter=True, print_matrices=(0, 0, 0, 0, 0)):
    """ Compute the electrical conductivity

    :param workspace: domain
    :type workspace: pumapy.Workspace
    :param cond_map: local constituents electrical conductivities
    :type cond_map: IsotropicConductivityMap or AnisotropicConductivityMap
    :param direction: direction for solve ('x','y', or 'z')
    :type direction: string
    :param side_bc: side boundary conditions can be symmetric ('s'), periodic ('p') or dirichlet ('d')
    :type side_bc: string
    :param prescribed_bc: 3D array holding dirichlet BC
    :type prescribed_bc: pumapy.ConductivityBC or None
    :param tolerance: tolerance for iterative solver
    :type tolerance: float
    :param maxiter: maximum Iterations for solver
    :type maxiter: int
    :param solver_type: solver type, options: 'bicgstab' (default), 'cg', 'gmres', 'direct'
    :type solver_type: string
    :param display_iter: display iterations and residual
    :type display_iter: bool
    :param print_matrices: corresponding to E, A, b, T, q decimal places. If 0, they are not printed
    :type print_matrices: (bool, bool, bool, bool, bool)
    :return: electrical conductivity, potential field, flux
    :rtype: ((float, float, float), numpy.ndarray, numpy.ndarray)

    :Example:
    >>> import pumapy as puma
    >>> ws_fiberform = puma.import_3Dtiff(puma.path_to_example_file("200_fiberform.tif"), 1.3e-6)
    >>> ws_fiberform.matrix = ws_fiberform.matrix[50:150, 50:150, 50:150]
    >>> cond_map = puma.IsotropicConductivityMap()
    >>> cond_map.add_material((0, 89), 0.0257)
    >>> cond_map.add_material((90, 255), 12)
    >>> k_eff_x, P_x, q_x = puma.compute_electrical_conductivity(ws_fiberform, cond_map, 'x', 's', tolerance=1e-2, solver_type='cg')
    """
    return compute_thermal_conductivity(workspace, cond_map, direction, side_bc, prescribed_bc, tolerance, maxiter,
                                        solver_type, display_iter, print_matrices)



import unittest
import numpy as np
from pumapy.utilities.isosurface import generate_isosurface
import pumapy as puma
from skimage import io


class TestIsosurface(unittest.TestCase):

    def test_fiberform(self):
        ws = puma.Workspace.from_array(io.imread(puma.path_to_example_file("200_fiberform.tif")))
        mesh = generate_isosurface(ws, (90, 255), True, False)

        np.testing.assert_equal(len(mesh.values), 665453)
        np.testing.assert_equal(mesh.faces.shape, (1330866, 3))


class TestSurfacearea(unittest.TestCase):

    def test_fiberform(self):
        ws = puma.Workspace.from_array(io.imread(puma.path_to_example_file("200_fiberform.tif")))
        area, specific_area = puma.compute_surface_area(ws, (90, 255), False)

        np.testing.assert_almost_equal(area, 4.12090e-7, decimal=5)
        np.testing.assert_almost_equal(specific_area, 52291.746, decimal=3)


class TestMeanInterceptLength(unittest.TestCase):

    def test_fiberform(self):
        ws = puma.import_3Dtiff(puma.path_to_example_file("200_fiberform.tif"), 1.)
        mil = puma.compute_mean_intercept_length(ws, (0, 89))
        np.testing.assert_almost_equal(mil, (63.26930970616208, 70.25351564671224, 54.3159003008165))


class TestVolumeFraction(unittest.TestCase):

    def test_fiberform(self):
        ws = puma.import_3Dtiff(puma.path_to_example_file("100_fiberform.tif"), 1.3e-6)
        vf = puma.compute_volume_fraction(ws, cutoff=(90, 255))
        np.testing.assert_almost_equal(vf, 0.16714)


class TestOrientationST(unittest.TestCase):

    def test_artfib(self):
        ws1 = puma.import_vti("testdata/artifib.vtk")
        puma.compute_orientation_st(ws1, (128, 255), 0.7, 1.1)
        ws2 = puma.import_vti("testdata/artifib_orient.vtk")
        error, mean, std = puma.compute_angular_differences(ws1.matrix, ws1.orientation, ws2.orientation, (128, 255))
        self.assertTrue(mean < 5 and std < 10)

    def test_artfib_edt(self):
        ws1 = puma.import_vti("testdata/artifib.vtk")
        puma.compute_orientation_st(ws1, (128, 255), 0.7, 1.1, edt=True)
        ws2 = puma.import_vti("testdata/artifib_orient.vtk")
        error, mean, std = puma.compute_angular_differences(ws1.matrix, ws1.orientation, ws2.orientation, (128, 255))
        print(mean, std)
        self.assertTrue(mean < 6 and std < 10)


if __name__ == '__main__':
    unittest.main()



from pumapy.utilities.workspace import Workspace
import numpy as np


class ConductivityBC:
    def __init__(self, workspace):
        """ Conductivity prescribed boundary conditions class. This contains two arrays:

            :param workspace: domain
            :type workspace: Workspace

            :var dirichlet: of size (X, Y, Z), it describes voxels to keep fixed to a certain temperature
            :vartype dirichlet: np.ndarray
        """
        self.dirichlet = np.full((workspace.len_x(), workspace.len_y(), workspace.len_z()), np.Inf, dtype=float)

    def show_dirichlet(self):
        Workspace.show_matrix(self.dirichlet)


class Isotropic_symmetricBC:
    def __init__(self, len_x, len_y, len_z):
        self.len_x = len_x
        self.len_y = len_y
        self.len_z = len_z
        self.len_xy = self.len_x * self.len_y

    def indexAt_3D(self, i, j, k):
        if i == -1:
            i = 0
        if i == self.len_x:
            i = self.len_x - 1
        if j == -1:
            j = 0
        if j == self.len_y:
            j = self.len_y - 1
        if k == -1:
            k = 0
        if k == self.len_z:
            k = self.len_z - 1
        return self.len_xy * k + self.len_x * j + i


class Isotropic_periodicBC:
    def __init__(self, len_x, len_y, len_z):
        self.len_x = len_x
        self.len_y = len_y
        self.len_z = len_z
        self.len_xy = self.len_x * self.len_y

    def indexAt_3D(self, i, j, k):
        if i == -1:
            i = self.len_x - 1
        if i == self.len_x:
            i = 0
        if j == -1:
            j = self.len_y - 1
        if j == self.len_y:
            j = 0
        if k == -1:
            k = self.len_z - 1
        if k == self.len_z:
            k = 0
        return self.len_xy * k + self.len_x * j + i


class ElasticityBC:
    def __init__(self, workspace):
        """ Elasticity prescribed boundary conditions class.
            N.B. links attribute takes precedence over dirichlet if a voxel prescribed for both attributes

            :param workspace: domain
            :type workspace: Workspace

            Attributes:
                dirichlet np.ndarray of size (X, Y, Z, 3), it describes voxels and DOFs to keep fixed to a certain
                          displacement. If unset, value is np.Inf, otherwise a float
                links:    np.ndarray of size (X, Y, Z, 3, 4), it describes voxels to link to each other,
                          provided as coordinates to link  e.g. bc.links[2, 4, 5, 2] = [0, 0, 0, 0] links the
                          z-component of voxel in position (2, 4, 5) to the x-component of voxel in position (0, 0, 0)
                indices   np.ndarray of size (X, Y, Z, 3, 4) containing the indices of the domain
        """
        self.dirichlet = np.full((workspace.len_x(), workspace.len_y(), workspace.len_z(), 3), np.Inf, dtype=float)
        self.links = np.full((workspace.len_x(), workspace.len_y(), workspace.len_z(), 3, 4), -1, dtype=int)
        self.indices = np.indices((workspace.len_x(), workspace.len_y(), workspace.len_z(), 3)).transpose((1, 2, 3, 4, 0))

    def show_dirichlet(self):
        Workspace.show_orientation(self.dirichlet)

    def show_links(self):
        Workspace.show_orientation(self.links)



import unittest
import numpy as np
import pumapy as puma


class TestSegmentation(unittest.TestCase):

    def test_identify_porespace(self):
        ws = puma.generate_sphere((100, 100, 100), (50, 50, 50), 40, segmented=False)
        ws_cp = ws.copy()
        ws_cp.binarize_range((1, 255))

        ws.binarize_range((1, 253))
        pores = puma.identify_porespace(ws, (1, 1))

        np.testing.assert_equal(np.where(pores == 0), np.where(ws.matrix == 1))
        np.testing.assert_equal(np.where(pores == 1, 1, 0).sum(), np.where(ws_cp.matrix == 0, 1, 0).sum())

    def test_fill_closed_pores(self):
        ws = puma.generate_random_spheres((100, 100, 100), diameter=20, porosity=0.8, allow_intersect=True, segmented=False)
        ws_cp = ws.copy()
        ws_cp.binarize_range((1, 250))

        ws.binarize_range((1, 250))
        filled_ws, pores = puma.fill_closed_pores(ws, (1, 1), fill_value=2, return_pores=True)

        np.testing.assert_equal(np.where(pores == 0), np.where(ws.matrix == 1))
        np.testing.assert_equal(np.where(pores >= 1, 1, 0).sum(), np.where(ws_cp.matrix == 0, 1, 0).sum())
        np.testing.assert_equal(np.where(pores >= 2), np.where(filled_ws.matrix == 2))


if __name__ == '__main__':
    unittest.main()



import unittest
import pumapy as puma
import numpy as np
np.random.seed(0)


class TestCylinderSquareArray(unittest.TestCase):

    def test_2d_square_array(self):
        ws = puma.generate_cylinder_square_array(300, 0.8, segmented=True)
        np.testing.assert_equal(ws.matrix.shape, (300, 300, 1))
        np.testing.assert_almost_equal(ws.porosity(), 0.8, 3)

        ws = puma.generate_cylinder_square_array(300, 0.8, segmented=False)
        np.testing.assert_equal(ws.matrix.shape, (300, 300, 1))
        np.testing.assert_almost_equal(ws.porosity(cutoff=(0, 127)), 0.8, 3)


class TestSphere(unittest.TestCase):

    def test_sphere(self):
        ws = puma.generate_sphere((100, 100, 100), (50, 50, 50), 80, segmented=True)
        analytical_porosity = (100. ** 3 - (4./3.) * np.pi * (40. ** 3)) / (100. ** 3)
        np.testing.assert_equal(ws.matrix.shape, (100, 100, 100))
        np.testing.assert_almost_equal(ws.porosity(), analytical_porosity, 4)

        ws = puma.generate_sphere((100, 100, 100), (50, 50, 50), 80, segmented=False)
        np.testing.assert_equal(ws.matrix.shape, (100, 100, 100))
        np.testing.assert_almost_equal(ws.porosity(cutoff=(0, 127)), analytical_porosity, 4)


class TestRandomSpheres(unittest.TestCase):

    def test_random_spheres(self):
        ws = puma.generate_random_spheres(shape=(200, 200, 200), diameter=4, porosity=0.75,
                                          allow_intersect=True, segmented=True)
        np.testing.assert_equal(ws.matrix.shape, (200, 200, 200))
        self.assertLess(abs(ws.porosity() - 0.75), 0.0001)

        ws = puma.generate_random_spheres(shape=(100, 100, 100), diameter=8, porosity=0.8,
                                          allow_intersect=True, segmented=False)
        np.testing.assert_equal(ws.matrix.shape, (100, 100, 100))
        self.assertLess(abs(ws.porosity(cutoff=(0, 127)) - 0.8), 0.00013)

        ws = puma.generate_random_spheres(shape=(50, 50, 50), diameter=4, porosity=0.8,
                                          allow_intersect=False, segmented=False)
        np.testing.assert_equal(ws.matrix.shape, (50, 50, 50))
        self.assertLess(abs(ws.porosity(cutoff=(0, 127)) - 0.8), 0.00023)


class TestRandomFibers(unittest.TestCase):

    def test_random_fibers(self):
        ws = puma.generate_random_fibers(shape=(100, 100, 100), radius=2, porosity=0.8, length=50, max_iter=6,
                                         allow_intersect=True, segmented=True)
        np.testing.assert_equal(ws.matrix.shape, (100, 100, 100))
        self.assertLess(abs(ws.porosity() - 0.8), 0.001)

        ws = puma.generate_random_fibers(shape=(100, 100, 100), radius=2, nfibers=100, phi=20, theta=90, length=100,
                                         max_iter=6, allow_intersect=True, segmented=True)
        np.testing.assert_equal(ws.matrix.shape, (100, 100, 100))
        self.assertEqual(ws.max(), 100)

        ws = puma.generate_random_fibers(shape=(50, 50, 50), radius=2, porosity=0.8, length=50, max_iter=6,
                                         allow_intersect=False, segmented=True)
        np.testing.assert_equal(ws.matrix.shape, (50, 50, 50))
        self.assertLess(abs(ws.porosity() - 0.8), 0.01)

        ws = puma.generate_random_fibers(shape=(100, 100, 100), radius=2, porosity=0.8, length=50, max_iter=6,
                                         allow_intersect=True, segmented=False)
        np.testing.assert_equal(ws.matrix.shape, (100, 100, 100))
        self.assertLess(abs(ws.porosity(cutoff=(0, 127)) - 0.8), 0.001)

    def test_51x44x87fibers_nfibers(self):
        nfibers = 100
        shape = (51, 44, 87)
        ws = puma.generate_random_fibers(shape=shape, radius=3, nfibers=nfibers, phi=20, theta=90, length=200,
                                         allow_intersect=True)

        self.assertEqual(ws.get_shape(), shape)
        self.assertEqual(ws.unique_values().shape[0], nfibers + 1)

    def test_fibers_porosity(self):
        porosity = 0.8
        ws = puma.generate_random_fibers(shape=(100, 100, 100), radius=3, porosity=porosity, phi=20, theta=90, length=200,
                                         allow_intersect=True, max_iter=20)

        self.assertTrue(abs(puma.compute_volume_fraction(ws, (0, 0)) - porosity) < 1e-2, 3)

    def test_fibers_nfibers_nointersection(self):
        nfibers = 100
        ws = puma.generate_random_fibers(shape=(100, 100, 100), radius=3, nfibers=nfibers, phi=20, theta=90, length=200,
                                         allow_intersect=False)

        self.assertEqual(ws.unique_values().shape[0], nfibers + 1)

    def test_fibers_porosity_nointersection(self):
        porosity = 0.9
        ws = puma.generate_random_fibers(shape=(100, 100, 100), radius=3, porosity=porosity, phi=20, theta=90, length=200,
                                         allow_intersect=False, max_iter=20)

        self.assertTrue(abs(puma.compute_volume_fraction(ws, (0, 0)) - porosity) < 1e-2, 3)


class TestTPMS(unittest.TestCase):

    def test_tpms(self):
        shape = (100, 100, 100)  # size of the domain, in voxels.
        w = 0.08  # value of w in the equations above
        q = 0.2  # value of q in the equations above
        ws = puma.generate_tpms(shape, w, q, equation=0)
        np.testing.assert_equal(ws.matrix.shape, (100, 100, 100))
        print(ws.porosity())

        ws = puma.generate_tpms(shape, w, q, equation=1)
        np.testing.assert_equal(ws.matrix.shape, (100, 100, 100))
        print(ws.porosity())

        ws = puma.generate_tpms(shape, w, q, equation=2)
        np.testing.assert_equal(ws.matrix.shape, (100, 100, 100))
        print(ws.porosity())


if __name__ == '__main__':
    unittest.main()
